
#include "PrecisionTime.h"
#include <Arduino.h>
#include <EEPROM.h>

// Manual for library: http://lygte-info.dk/project/PrecisionTime%20UK.html
// This is version 1.00 from 2019-10-2
// By HKJ from lygte-info.dk


char PrecisionTimeId[4] = "PTt"; // Identifier for calibration value

PrecisionTime::PrecisionTime(unsigned int eepromAddress) {
  this->eepromAddress = eepromAddress;
  char idcheck[sizeof(PrecisionTimeId)];
  EEPROM.get(eepromAddress, idcheck);
  if (strcmp(idcheck, PrecisionTimeId) == 0) {
    EEPROM.get(eepromAddress + sizeof(PrecisionTimeId), oneSecondCountAdjustment);
  }
}

byte PrecisionTime::getDeltaSeconds() {
  noInterrupts();
  byte d = oneSecond;
  oneSecond = 0;
  interrupts();
  return d;
}


void PrecisionTime::startCalibration(unsigned long baud) {
  Serial.println(F("StartCalibration"));
  delay(100);
  Serial.begin(76800);  // This has very low error making it a very reliable baudrate
  delay(500);
  Serial.println(F("StartCalibration fast"));
  boolean done = false;
  long dt = 0;
#if defined(_uno_) || defined(_mega_) || defined(_U4_)
  TIMSK0 &= ~(1 << TOIE0); // Disable millis interrupt
#endif
  while (!done) {
    dt += getDeltaSeconds();
    if (Serial.available() > 0) {
      switch (Serial.read()) {
        case '?' : while (oneSecond == 0); Serial.println(dt + 1); break;
        case '!' : noInterrupts(); dt = 0; rtcCounter = 0; oneSecond = 0; interrupts(); break;
        case 'c' : calibrate(); break;
        case 'f' : Serial.println(oneSecondCountAdjustment); break;
        case 'q' : done = true; break;
      }
    }
  }
#if defined(_uno_) || defined(_mega_) || defined(_U4_)
  TIMSK0 |= (1 << TOIE0); // Enable millis interrupt
#endif
  Serial.begin(baud);
}

void PrecisionTime::startMillisCheck(unsigned long baud) {
  Serial.println(F("StartCalibration"));
  delay(100);
  Serial.begin(76800);  // This has very low error making it a very reliable baudrate
  delay(500);
  Serial.println(F("StartCalibration fast"));
  boolean done = false;
  while (!done) {
    if (Serial.available() > 0) {
      switch (Serial.read()) {
        case '?' : {
            unsigned v = (unsigned) (millis() / 1000);
            while (v == (unsigned) (millis() / 1000));
            Serial.println(v + 1);
          }
          break;
        case 'f' : Serial.print('X'); Serial.println(oneSecondCountAdjustment); break;
        case 'q' : done = true; break;
      }
    }
  }
  Serial.begin(baud);
}

void PrecisionTime::updateIntr() {
  rtcCounter += rtcUpdate;
  if (rtcCounter >= refOneSecondCount) {
    rtcCounter -= (refOneSecondCount + oneSecondCountAdjustment);
    oneSecond++;
  }
}

void PrecisionTime::clearCalibration() {
  EEPROM.put(eepromAddress, PrecisionTimeId);
  EEPROM.put(eepromAddress + sizeof(PrecisionTimeId), 0);
}

void PrecisionTime::calibrate() {
  char c;
  boolean s = false;
  long n = 0;
  do {
    while (Serial.available() == 0);
    c = Serial.read();
    if (c == '-') s = true;
    else if (c >= '0' && c <= '9') n = n * 10 + (c - '0');
  } while (c != '\n');
  if (s) n = -n;
  oneSecondCountAdjustment = n;
  EEPROM.put(eepromAddress, PrecisionTimeId);
  EEPROM.put(eepromAddress + sizeof(PrecisionTimeId), oneSecondCountAdjustment);
}

#ifdef _INCLUDE_TIME_
const PROGMEM byte daysInMonthList[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

void PrecisionTime::update() {
  time.increment(getDeltaSeconds());
}

byte Time::daysInMonth() {
  if (month == 1 && ((year % 4 == 0 && !year % 100 == 0) || year % 400 == 0)) return 29;
  return daysInMonthList[month];
}

void Time::increment(byte increment) {
  while (increment > 0) {
    increment--;
    second++;
    if (second >= 60) {
      second -= 60;
      minute++;
      if (minute >= 60) {
        minute -= 60;
        hour++;
        if (hour >= 24) {
          hour -= 24;
          day++;
          if (day > daysInMonth()) {
            day -= daysInMonth();
            month++;
            if (month >= 12) {
              month -= 12;
              year++;
            }
          }
        }
      }
    }

  }
}
#endif

//******************************************************************************************************
//******************************************************************************************************
// Init interrupt
// All routines setup a 1kHz interrupt
//******************************************************************************************************
//******************************************************************************************************

#if defined(_USE_TIMER0_)
#if defined(_mega0_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCB0.CCMP = F_CPU / 1000; // 1000 intr/sec
  TCB0.CTRLA = 1; // Enable timer
  TCB0.CTRLB = 0; // Periodic interrupt mode
  TCB0.INTCTRL = 1; // Enable interrupt
  interrupts();
}
#else
#error "Interrupt 1 not supported on this processor"
#endif
#elif defined(_USE_TIMER1_)
#if defined(_uno_) || defined(_mega_) || defined(_U4_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCCR1A = 0;
  TCCR1B = 0;
  OCR1A = (F_CPU / 1000) - 1; // set compare match register for 1khz
  TCCR1B = (1 << CS10) | (1 << WGM12); // Set for /1 prescaler and CTC mode
  TIMSK1 = (1 << OCIE1A); // enable timer compare interrupt
  interrupts();
}
#elif defined(_mega0_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCB1.CCMP = F_CPU / 1000; // 1000 intr/sec
  TCB1.CTRLA = 1; // Enable timer
  TCB1.CTRLB = 0; // Periodic interrupt mode
  TCB1.INTCTRL = 1; // Enable interrupt
  interrupts();
}
#else
#error "Interrupt 1 not supported on this processor"
#endif
#elif defined(_USE_TIMER2_)
#if defined(_uno_) || defined(_mega_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCCR2A = 0;
  TCCR2B = 0;
  OCR2A = (F_CPU / 64000) - 1; // set compare match register for 1khz
  TCCR2A = (1 << WGM21);  // turn on CTC mode
  TCCR2B = (1 << CS22); // Set for /64 prescaler
  TIMSK2 = (1 << OCIE2A); // enable timer compare interrupt
  interrupts();
}
#elif defined(_mega0_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCB2.CCMP = F_CPU / 1000; // 1000 intr/sec
  TCB2.CTRLA = 1; // Enable timer
  TCB2.CTRLB = 0; // Periodic interrupt mode
  TCB2.INTCTRL = 1; // Enable interrupt
  interrupts();
}
#else
#error "Interrupt 2 not supported on this processor"
#endif
#elif defined(_USE_TIMER3_)
#if defined(_mega_) || defined(_U4_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCCR3A = 0;
  TCCR3B = 0;
  OCR3A = (F_CPU / 1000) - 1; // set compare match register for 1khz
  TCCR3B = (1 << CS30) | (1 << WGM32); // Set for /1 prescaler and CTC mode
  TIMSK3 = (1 << OCIE3A); // enable timer compare interrupt
  interrupts();
}
#elif defined(_mega0_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCB3.CCMP = F_CPU / 1000; // 1000 intr/sec
  TCB3.CTRLA = 1; // Enable timer
  TCB3.CTRLB = 0; // Periodic interrupt mode
  TCB3.INTCTRL = 1; // Enable interrupt
  interrupts();
}
#else
#error "Interrupt 3 not supported on this processor"
#endif
#elif defined(_USE_TIMER4_)
#if defined(_U4_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  OCR4C = (F_CPU / 64000); // set compare match register for 1khz
  TCCR4A = 0;
  TCCR4B = (1 << CS42) | (1 << CS41); // Prescaler is /32
  TIMSK4 = (1 << TOIE4);
  interrupts();
}
#elif defined(_mega_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCCR4A = 0;
  TCCR4B = 0;
  OCR4A = (F_CPU / 1000) - 1; // set compare match register for 1khz
  TCCR4B = (1 << CS40) | (1 << WGM42); // Set for /1 prescaler and CTC mode
  TIMSK4 = (1 << OCIE4A); // enable timer compare interrupt
  interrupts();
}
#else
#error "Interrupt 4 not supported on this processor"
#endif
#elif defined(_USE_TIMER5_)
#if defined(_mega_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCCR5A = 0;
  TCCR5B = 0;
  OCR5A = (F_CPU / 1000) - 1; // set compare match register for 1khz
  TCCR5B = (1 << CS50) | (1 << WGM52); // Set for /1 prescaler and CTC mode
  TIMSK5 = (1 << OCIE5A); // enable timer compare interrupt
  interrupts();
}
#elif defined(_mega0_)
void PrecisionTime::initInterrupt() {
  noInterrupts();
  TCA0.SINGLE.PER = F_CPU / 1000; // 1000 intr/sec
  TCA0.SINGLE.CTRLA = 1; // Enable timer at clock freq
  TCA0.SINGLE.CTRLB = 0; // Normal mode
  TCA0.SINGLE.INTCTRL = 1; // Enable interrupt
  interrupts();
}
#else
#error "Interrupt 5 not supported on this processor"
#endif

#else
void PrecisionTime::initInterrupt() {
}
#endif
