
#ifndef __PRECISIONTIME_
#define __PRECISIONTIME_

#include <arduino.h>
#include <Time.h>

// Manual for library: http://lygte-info.dk/project/PrecisionTime%20UK.html
// This is version 1.00 from 2019-10-2
// By HKJ from lygte-info.dk


// Define what timer to use for interrupt
//#define _USE_TIMER0_   // Supported on Mega4809(B0)
#define _USE_TIMER1_   // Supported on Mega328 & Mega2560 & Mega32U4 & Mega4809(B1)
//#define _USE_TIMER2_   // Supported on Mega328 & Mega2560 & Mega4809(B2)
//#define _USE_TIMER3_   // Supported on Mega2560 & Mega32U4
//#define _USE_TIMER4_   // Supported on Mega2560 & Mega32U4
//#define _USE_TIMER5_   // Supported on Mega2560

#define _INCLUDE_TIME_    // Include a time struct to hand seconds,minutes,hours,date,month,year, this is 7 bytes

#define PRECISIONTIME_EEPROM_ADDRESS 0x78 // Needs 8 bytes



const long refOneSecondCount = 24L * 60L * 60L * 100L;    // 0.01 second in a day resolution
const long rtcUpdate = refOneSecondCount / 1000;  // With 1000 interrupts a second this value must be /1000 of the above

// A time class with bytes value for most compact size
class Time {
  private:
    friend class PrecisionTime;
    byte second;
    byte minute;
    byte hour;
    byte day;
    byte month;
    unsigned int year;

    void clear() {
      second = 0;
      month = 0;
      hour = 0;
      day = 0;
      month = 0;
      year = 0;
    }
  public:
    Time() {
      clear();
    }
    void increment(byte seconds); // Add some seconds to time
    byte daysInMonth(); // Return days in current month
    byte getSecond() {
      return second;
    }
    byte getMinute() {
      return minute;
    }
    byte getHour() {
      return hour;
    }
    byte getDay() {
      return day + 1;
    }
    byte getMonth() {
      return month + 1;
    }
    byte getYear() {
      return year;
    }
};

class PrecisionTime {
  private:
    long oneSecondCountAdjustment = 0; // Calibration factor
    unsigned int eepromAddress;
    volatile byte oneSecond = 0;      // This value is incremented each second
#ifdef _INCLUDE_TIME_
    Time time;                        // Compact structure to store data and time in.
#endif
    void calibrate();                 // Change the calibration factor with value from Serial
    void initInterrupt();
    long rtcCounter = 0;

  public:
    PrecisionTime(unsigned int eepromAddress);
    void begin() {                    // Start the interrupt
      initInterrupt();
    }

    void updateIntr();  // Must be called from a 1ms or 1kHz interrupt

    byte getDeltaSeconds();           // Get accumulated seconds since last call
    boolean isCalibrated() {
      return oneSecondCountAdjustment!=0;
    }
    void startCalibration() {         // Start calibration, assume 9600 baud
      startCalibration(9600);
    }
    void startMillisCheck() {         // Check standard Arduino precision, assume 9600 baud
      startMillisCheck(9600);
    }

    void clearCalibration();          // Set calibration to zero

    void startCalibration(unsigned long baud);  // Start calibration
    void startMillisCheck(unsigned long baud);  // Check standard Arduino precision

#ifdef _INCLUDE_TIME_
    void update();                    // Transfer delta seconds to the time counter
    void SetTime(unsigned int year, byte month, byte day, byte hour, byte minute, byte second) {
      time.year = year;
      time.month = month > 0 ? month - 1 : 0;
      time.day = day > 0 ? day - 1 : 0;
      time.hour = hour;
      time.minute = minute;
      time.second = second;
    }

    Time getTime() {
      return time;
    }
    byte getSecond() {
      return time.second;
    }
    byte getMinute() {
      return time.minute;
    }
    byte getHour() {
      return time.hour;
    }
    byte getDay() {
      return time.day + 1;
    }
    byte getMonth() {
      return time.month + 1;
    }
    int getYear() {
      return time.year;
    }
    void setSecond(byte second) {
      time.second = second;
    }
    void setMinute(byte minute) {
      time.minute = minute;
    }
    void setHour(byte hour) {
      time.hour = hour;
    }
    void setDay(byte day) {
      time.day = day > 0 ? day - 1 : 0;
    }
    void setMonth(byte month) {
      time.month = month > 0 ? month - 1 : 0;
    }
    void setYear(int year) {
      time.year = year;
    }
    // Convert the internal time structure to standard C types, this uses a lot of flash memory
    time_t getTM(struct tm * tmStruct) {
      tmStruct->tm_sec = time.second;
      tmStruct->tm_min = time.minute;
      tmStruct->tm_hour = time.hour;
      tmStruct->tm_mday = time.day + 1;
      tmStruct->tm_mon = time.month + 1;
      tmStruct->tm_year = time.year;
      if (tmStruct->tm_year >= 1900) tmStruct->tm_year -= 1900;
      return mktime(tmStruct);
    }
#endif
};

//******************************************************************************************************
//******************************************************************************************************
// Categorize the different Atmel processors, this is used for timer interrupt handling
//******************************************************************************************************
//******************************************************************************************************

#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega168P__) || defined(__AVR_ATmega328__) || defined(__AVR_ATmega328P__)
#define _uno_
#elif defined(__AVR_ATmega2560__) || defined(__AVR_ATmega1280__)
#define _mega_
#elif defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega16U4__)
#define _U4_
#elif defined(__AVR_ATmega4809__) || defined(__AVR_ATmega4808__)
#define _mega0_
#endif


//******************************************************************************************************
//******************************************************************************************************
// Define interrupt vector, this macro is empty when no interrupt is defined
//******************************************************************************************************
//******************************************************************************************************

#if defined(_USE_TIMER0_)
#if defined(_mega0_)
#define PRECISION_TIMER_INTR(precisionTime) ISR(TCB0_INT_vect) { TCB0.INTFLAGS=1;precisionTime.updateIntr();}
#endif
#endif

#if defined(_USE_TIMER1_)
#if defined(_mega0_)
#define PRECISION_TIMER_INTR(precisionTime) ISR(TCB1_INT_vect) { TCB1.INTFLAGS=1;precisionTime.updateIntr();}
#else
#define PRECISION_TIMER_INTR(precisionTime) ISR(TIMER1_COMPA_vect) { precisionTime.updateIntr();}
#endif
#endif

#if defined(_USE_TIMER2_)
#if defined(_mega0_)
#define PRECISION_TIMER_INTR(precisionTime) ISR(TCB2_INT_vect) { TCB2.INTFLAGS=1;precisionTime.updateIntr();}
#else
#define PRECISION_TIMER_INTR(precisionTime) ISR(TIMER2_COMPA_vect) { precisionTime.updateIntr();}
#endif
#endif

#if defined(_USE_TIMER3_)
#if defined(_mega0_)
#define PRECISION_TIMER_INTR(precisionTime) ISR(TCB3_INT_vect) { TCB3.INTFLAGS=1;precisionTime.updateIntr();}
#else
#define PRECISION_TIMER_INTR(precisionTime) ISR(TIMER3_COMPA_vect) { precisionTime.updateIntr();}
#endif
#endif

#if defined(_USE_TIMER4_)
#if defined(_U4_)
#define PRECISION_TIMER_INTR(precisionTime) ISR(TIMER4_OVF_vect) { precisionTime.updateIntr();}
#elif defined(_mega_)
#define PRECISION_TIMER_INTR(precisionTime) ISR(TIMER4_COMPA_vect) { precisionTime.updateIntr();}
#endif
#endif

#if defined(_USE_TIMER5_)
#if defined(_mega0_)
#define PRECISION_TIMER_INTR(precisionTime) ISR(TCA0_OVF_vect) { TCA0.SINGLE.INTFLAGS=1;precisionTime.updateIntr();}
#elif defined(_mega_)
#define PRECISION_TIMER_INTR(precisionTime) ISR(TIMER5_COMPA_vect) { precisionTime.updateIntr();}
#endif
#endif

#ifndef PRECISION_TIMER_INTR
#define PRECISION_TIMER_INTR(precisionTime)
#endif


#endif
